<?php
/*
 * =============================================================================
 * Query Source Engine servers for details, map, players, etc
 * 
 * @copyright SourceBans (C)2007 SteamFriends.com.  All rights reserved.
 * @package SourceBans
 * @link http://www.sourcebans.net
 * 
 * Edit by Dionys for PHP4 compatibility
 * Version 1.0.1 (2011-07-19)
 * skiner@inbox.ru
 * =============================================================================
*/

/** Constants of the data we need for a query */
define('CSERVERINFO_QUERY_HEADER',							"\xFF"."\xFF"."\xFF"."\xFF");
define('CSERVERINFO_A2S_INFO',								"\x54"."Source Engine Query"."\0");
define('CSERVERINFO_A2S_INFO_RESPONSE',						"\x49");
define('CSERVERINFO_A2S_INFO_RESPONSE_OLD',					"\x6D");
define('CSERVERINFO_A2A_PING',								"\x69");	// Deprecated
define('CSERVERINFO_A2A_PING_RESPONSE',						"\x6A");	// Deprecated
define('CSERVERINFO_A2S_PLAYER',							"\x55");
define('CSERVERINFO_A2S_PLAYER_RESPONSE',					"\x44");
define('CSERVERINFO_A2S_SERVERQUERY_GETCHALLENGE',			"\x57");
define('CSERVERINFO_A2S_SERVERQUERY_GETCHALLENGE_RESPONSE',	"\x41");
define('CSERVERINFO_A2S_RULES',								"\x56");
define('CSERVERINFO_A2S_RULES_RESPONSE',					"\x45");

class CServerInfo
{
	var $raw;
	var $address;
	var $port;
	var $socket		= false;
	var $isfsock	= true;
	var $bindip		= "";

	function CServerInfo($address, $port)
	{
		$this->address		= $address;
		$this->port			= $port;
		$this->challenge	= CSERVERINFO_QUERY_HEADER;
	}

	function getInfo()
	{
		$ret	= array();
		$socket	= $this->_getSocket();
		$mT1	= mTime();	// Check start Request ping time
		$packet	= $this->_request($socket, CSERVERINFO_A2S_INFO);
		$mT2	= mTime();	// Check end Request ping time
		$pT		= ceil((($mT2 - $mT1) * 1000)/2);	//Calc ping

		if(empty($packet))
		{
			return $ret;
		}

		$ret['ping']	= $pT .' ms';
		$this->raw		= $packet;
		$type			= $this->_getraw(1);

		if($type == CSERVERINFO_A2S_INFO_RESPONSE)
		{
			// New protocol for Source and Goldsrc
			$ret['version']		= $this->_getbyte();
			$ret['hostname']	= $this->_getnullstr();	
			$ret['map']			= $this->_getnullstr();
			$ret['gamename']	= $this->_getnullstr();
			$ret['gamedesc']	= $this->_getnullstr();
			$this->_getushort();	// AppId
			$ret['numplayers']	= $this->_getbyte();
			$ret['maxplayers']	= $this->_getbyte();
			$ret['botcount']	= $this->_getbyte();
			$ret['dedicated']	= $this->_getraw(1);
			$ret['os']			= $this->_getraw(1);
			$ret['password']	= $this->_getbyte();
			$ret['secure']		= $this->_getbyte();
		}
		else if($type == CSERVERINFO_A2S_INFO_RESPONSE_OLD)
		{
			// Legacy Goldsrc support
			$this->_getnullstr();	// GameIP
			$ret['hostname']	= $this->_getnullstr();
			$ret['map']			= $this->_getnullstr();
			$ret['gamename']	= $this->_getnullstr();
			$ret['gamedesc']	= $this->_getnullstr();
			$ret['numplayers']	= $this->_getbyte();
			$ret['maxplayers']	= $this->_getbyte();
			$ret['version']		= $this->_getbyte();
			$ret['dedicated']	= $this->_getraw(1);
			$ret['os']			= $this->_getraw(1);
			$ret['password']	= $this->_getbyte();

			if($this->_getbyte())	// IsMod
			{
				$this->_getnullstr();
				$this->_getnullstr();
				$this->_getbyte();
				$this->_getlong();
				$this->_getlong();
				$this->_getbyte();
				$this->_getbyte();
			}

			$ret['secure']		= $this->_getbyte();
			$ret['botcount']	= $this->_getbyte();
		}
		else
		{
			//TODO:
		}

		return $ret;				
	}
	
	function getPlayers()
	{
		$socket	= $this->_getSocket();
		$packet	= $this->_requestWithChallenge($socket, CSERVERINFO_A2S_PLAYER, CSERVERINFO_A2S_PLAYER_RESPONSE);

		if(empty($packet))
		{
			return array();
		}
		
		$this->raw	= $packet;
		$count		= $this->_getbyte();	 
		$players	= array();

		for($i = 0; $i < $count; $i++)
		{
			// Left4Dead2 returns an index of zero for all players. Need the index for the server context menu...
			$index = $this->_getbyte();

			if($index == 0)
			{
				$index = $i;
			}

			$temp =	array(
						'index'	=> $index,
						'name'	=> $this->_getnullstr(),
						'kills'	=> $this->_getlong(),
						'time'	=> secstostr(intval($this->_getfloat()))
					);

			if(!empty($temp['name']))
			{
				$players[] = $temp;
			}
		}

		array_qsort($players, 'kills', SORT_DESC);

		return $players;
	}
	
	function getRules()
	{
		$socket = $this->_getSocket();
		$packet = $this->_requestWithChallenge($socket, CSERVERINFO_A2S_RULES, CSERVERINFO_A2S_RULES_RESPONSE);

		if(empty($packet))
		{
			return array();
		}
		
		$this->raw	= $packet;
		$nump		= $this->_getushort(); 
		$ret		= array();
	
		for($i = 0; $i < $nump; $i++)
		{
			$name	= $this->_getnullstr();
			$value = $this->_getnullstr();
			
			if(!empty($name))
				$ret[$name] = $value;
		}
		
		ksort($ret);
		return $ret;
	}

	function _getSocket()
	{
		if($this->socket !== false)
		{
			return $this->socket;
		}

		if (!empty($this->bindip) && function_exists('socket_create') && function_exists('socket_bind'))
		{
			$this->isfsock	= false;
			$this->socket	= socket_create(AF_INET, SOCK_DGRAM, SOL_UDP);

			socket_set_option($this->socket, SOL_SOCKET, SO_REUSEADDR, 1);
			socket_bind($this->socket, $this->bindip);
			socket_connect($this->socket, $this->address, $this->port);
			socket_set_option($this->socket, SOL_SOCKET, SO_SNDTIMEO, array("sec"=>1, "usec"=>0));
			socket_set_option($this->socket, SOL_SOCKET, SO_RCVTIMEO, array("sec"=>1, "usec"=>0));
		}
		else
		{
			$this->socket	= @fsockopen("udp://".$this->address, $this->port, $errno, $errstr, 2);
			stream_set_timeout($this->socket, 1);
		}

		if($this->socket === false)
		{
			return false;
		}
		else
		{
			return $this->socket;
		}
	}

	function _request($socket, $code, $reply = null)
	{
		if ($this->isfsock)
		{
			fwrite($socket, CSERVERINFO_QUERY_HEADER. $code);
		}
		else
		{
			socket_write($socket, CSERVERINFO_QUERY_HEADER. $code, strlen(CSERVERINFO_QUERY_HEADER. $code));
		}

		$packet = $this->_readsplit($socket);
		if(empty($packet))
		{
			return "";
		}

		$this->raw = $packet;
		$magic = $this->_getlong();
		if($magic != -1)
		{
			return "";
		}

		$response = $this->_getraw(1);

		if($reply == null)
		{
			return substr($packet, 4);	// Skip magic as it was checked
		}
		else if($response == $reply)
		{
			return substr($packet, 5);	// Skip magic and type as it was checked
		}

		return "";
	}
	
	function _requestWithChallenge($socket, $code, $reply = null)
	{
		$maxretries = 5;
		while(true)
		{
			if ($this->isfsock)
			{
				fwrite($socket, CSERVERINFO_QUERY_HEADER. $code . $this->challenge); // do the request with challenge id = -1
			}
			else
			{
				socket_write($socket, CSERVERINFO_QUERY_HEADER. $code . $this->challenge, strlen(CSERVERINFO_QUERY_HEADER. $code . $this->challenge));
			}

			$packet = $this->_readsplit($socket);
			if(empty($packet))
			{
				return "";
			}
			
			$this->raw = $packet;
			$magic = $this->_getlong();
			if($magic != -1)
			{
				return "";
			}
			
			$response = $this->_getraw(1);
			if($response == CSERVERINFO_A2S_SERVERQUERY_GETCHALLENGE_RESPONSE)
			{
				$this->challenge = $this->_getraw(4);
			}
			else if($reply == null)
			{
				return substr($packet, 4);	// Skip magic as it was checked
			}
			else if($response == $reply)
			{
				return substr($packet, 5);	// Skip magic and type as it was checked
			}

			$maxretries --;

			if ($maxretries <= 0)
			{
				break;
			}
		}

		return "";
	}
	
	function _readsplit($socket)
	{
		if ($this->isfsock)
		{
			$packet = fread($socket, 1480);
		}
		else
		{
			$packet = socket_read($socket, 1480);
		}

		if(empty($packet))
		{
			return "";
		}
		
		$this->raw	= $packet;
		$type		= $this->_getlong();
		if($type == -2)
		{
			// Parse first header
			$reqid		= $this->_getlong();
			$packets	= $this->_getushort();
			$numpackets	= $packets & 0xFF;
			$curpacket	= $packets >> 8;

			if($reqid >= 0)	// Dummy value telling how big the split is (hardcoded to 1248), Orangebox or later
			{
				$this->_skip(2);
			}

			$data	= array();
			$tstart	= microtime(true);
			
			// Sanity
			if($curpacket >= $numpackets)
			{
				return "";
			}
			
			// Compressed?
			if($curpacket == 0 && $reqid < 0)
			{
				$sizeuncompressed	= $this->_getlong();
				$crc				= $this->_getlong();
			}
			
			while(true)
			{
				// Split already received (duplicate)?
				if(!array_key_exists($curpacket, $data))
				{
					$data[$curpacket] = $this->raw;
				}

				// Finished?
				if(count($data) >= $numpackets)
				{
					// Join the parts
					ksort($data);
					$data = implode("", $data);

					// Uncompress if necessary
					if($reqid < 0)
					{
						$data = bzdecompress($data);

						if(strlen($data) != $sizeuncompressed)
						{
							return "";
						}

						// TODO: CRC32 check
						return $data;
					}
					
					// Not compressed
					return $data;
				}

				// Check the timeout over several receives
				if(microtime(true) - $tstart >= 2.0)	// 2s
				{
					return "";
				}
				
				// Receive next packet
				if ($this->isfsock)
				{
					$packet = fread($socket, 1480);
				}
				else
				{
					$packet = socket_read($socket, 1480);
				}

				if(empty($packet))
				{
					return "";
				}

				// Parse packet
				$this->raw	= $packet;
				$_type		= $this->_getlong();
				if($_type != -2)
				{
					return "";
				}

				$_reqid			= $this->_getlong();
				$_packets		= $this->_getushort();
				$_numpackets	= $_packets & 0xFF;
				$curpacket		= $_packets >> 8;

				if($reqid >= 0)	// Dummy value telling how big the split is (hardcoded to 1248), Orangebox or later
				{
					$this->_skip(2);
				}

				// Sanity check
				if($_reqid != $reqid || $_numpackets != $numpackets || $curpacket >= $numpackets)
				{
					return "";
				}

				// Compressed?
				if($curpacket == 0 && $reqid < 0)
				{
					$sizeuncompressed	= $this->_getlong();
					$crc				= $this->_getlong();
				}
			}	
		}
		else if($type == -1)
		{
			// Non-split packet
			return $packet;
		}
		else
		{
			// Invalid
			return "";
		}
	}

	function _getraw($count)
	{
		$data		= substr($this->raw, 0, $count);
		$this->raw	= substr($this->raw, $count);
		return $data;
	}

	function _getbyte()
	{
		$byte	= $this->_getraw(1);
		return ord($byte);
	}

	function _getfloat()
	{
		$f	= @unpack('f1float', $this->_getraw(4));
		return $f['float'];
	}

	function _getlong()
	{
		$lo		= $this->_getushort();
		$hi		= $this->_getushort();
		$long	= ($hi << 16) | $lo;
		if ($long & 0x80000000 && $long > 0)	// This is special for register size >32 bits
		{
			return -((~$long & 0xFFFFFFFF) + 1);
		}
		else
		{
			return $long;						// 32-bit handles negative values implicitly
		}
	}

	function _getnullstr()
	{
		if (empty($this->raw))
		{
			return '';
		}

		$end		= strpos($this->raw, "\0");
		$str		= substr($this->raw, 0, $end);
		$this->raw	= substr($this->raw, $end + 1);

		return $str;
	}

	function _getushort()
	{
		$lo		= $this->_getbyte();
		$hi		= $this->_getbyte();
		$short	= ($hi << 8) | $lo;

		return $short;
	}
	
	function _getshort()
	{
		$short = $this->_getushort();

		if ($short & 0x8000)
		{
			return -((~$short & 0xFFFF) + 1);
		}
		else
		{
			return $short;
		}
	}

	function _skip($c)
	{
		$this->raw = substr($this->raw, $c);
	}
}

function mTime()
{ 
    list($msec, $sec) = explode(" ", microtime()); 
    return ((float)$msec + (float)$sec); 
}

function secstostr($sec, $textual = true)
{
	if($textual)
	{
		$div = array( 2592000, 604800, 86400, 3600, 60, 1 );
		$desc = array('mo','wk','d','hr','min','sec');
		$ret = null;
		foreach($div as $index => $value)
		{
			$quotent = floor($sec / $value); //greatest whole integer
			if($quotent > 0) {
				$ret .= "$quotent {$desc[$index]}, ";
				$sec %= $value;
			}
		}
		return substr($ret,0,-2);
	}
	else
	{
		$hours = floor ($sec / 3600);
		$sec -= $hours * 3600;
		$mins = floor ($sec / 60);
		$secs = $sec % 60;
		return "$hours:$mins:$secs";
	}
}

function array_qsort(&$array , $column = 0 , $order = SORT_ASC , $first = 0 , $last = - 2)
{
	// $array - the array to be sorted
	// $column - index (column) on which to sort
	// can be a string if using an associative array
	// $order - SORT_ASC (default) for ascending or SORT_DESC for descending
	// $first - start index (row) for partial array sort
	// $last - stop index (row) for partial array sort
	// $keys - array of key values for hash array sort

	$keys = array_keys ($array);

	if ($last == - 2)
	{
		$last = count ($array) - 1;
	}
	if ($last > $first)
	{
		$alpha = $first;
		$omega = $last;
		$key_alpha = $keys[$alpha];
		$key_omega = $keys[$omega];
		$guess = $array[$key_alpha][$column];

		while ($omega >= $alpha)
		{
			if ($order == SORT_ASC)
			{
				while ($array[$key_alpha][$column] < $guess)
				{
					$alpha ++;
					$key_alpha = $keys[$alpha];
				}
				while ($array[$key_omega][$column] > $guess)
				{
					$omega --;
					$key_omega = $keys[$omega];
				}
			}
			else
			{
				while ($array[$key_alpha][$column] > $guess)
				{
					$alpha ++;
					$key_alpha = $keys[$alpha];
				}
				while ($array[$key_omega][$column] < $guess)
				{
					$omega --;
					$key_omega = $keys[$omega];
				}
			}

			if ($alpha > $omega)
			{
				break;
			}

			$temporary = $array[$key_alpha];
			$array[$key_alpha] = $array[$key_omega];
			$alpha ++;

			$key_alpha = $keys[$alpha];
			$array[$key_omega] = $temporary;
			$omega --;

			$key_omega = $keys[$omega];
		}

		array_qsort ($array, $column, $order, $first, $omega);
		array_qsort ($array, $column, $order, $alpha, $last);
	}
}
?>